/*************************************************************
 * Philips USB peripheral controller driver
 *
 * (c) 2002 Koninklijke Philips Electronics N.V., All rights reserved
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * File Name:   usb_pdc.c
 *
 * Refering linux kernel version 2.6.9
 *
 * History:     
 *
 * Date            Author              Comments
 * -----------------------------------------------------------
 * Aug 9, 2004     SRIRAL PAULRAJ     Initial Creation
 * Sept 2, 2004     SRIRAM PAULRAJ    Modification for 1761 register access
 *
 *************************************************************/

/* Philips Peripheral Controller Driver
 *
 * The mapping of the endpoint is simple: it basically behaves as a device,
 * which can be opened, closed, and have reads and writes performed
 * on it. Due to the high bandwidth of the USB (12Mbit) we maintain local
 * buffers to ensure that we don't get starved of data during transmissions -
 * and have a receive buffer to allow the process dealing with USB to read in
 * bigger blocks than the packet size.
 *
 * The implementation is designed to be pretty transparent: this is for a
 * number of reasons, this is same protocol over both the USB and the serial 
 * ports on the customers unit.Implementing the endpoint as a simple 'open/close' device
 * as opposed to a more complex network-style interface also means that we can
 * do froody stuff like run PPP over a 12Mbit usb link (host permitting, of
 * course...). To this end, there is limited control over the way the USB
 * device works - endpoint 0 is handled totally under software control, and
 * only a limited number of events are passed though for the user-side task
 * to worry about (like connection/disconnection of the USB cable).
 *
 */

#ifdef LINUX_269
#include <linux/config.h>
#endif
#include <linux/types.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/major.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/init.h>
#include <linux/usb.h> 

#include <asm/byteorder.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>

#include "pdc_intf.h"
#include "usb_pdc.h"

#include "../hal/hal_intf.h"    
#include "../hal/isp1761.h"     

extern void pio_debug_write(int iValue);

/*--------------------------------------------------------------*
 *               external functions
 *--------------------------------------------------------------*/
extern  int     pdc_bus_init(struct pdc_dev     *pdc);
extern  void    pdc_bus_deinit(void);


/*--------------------------------------------------------------*
 *          Device Controller endpoint related functions
 *--------------------------------------------------------------*/
static void writeendpoint(int endpoint, u8 *buffer, int length);
static __inline__ int checkendpoint(int endpoint);
static void pdc_configure_eps(struct pdc_dev    *pdc);
static  int pdc_submit_control_urb(struct pdc_dev       *pdc, struct pdc_urb *urb_req);

/*--------------------------------------------------------------*
 *         Interrupt Service Routine related functions
 *--------------------------------------------------------------*/
void pdc_isr(struct isp1761_dev *dev, void *isr_data);
static void tx_data(struct pdc_dev *pdc, __u8 ep, int kick);
static void rx_data(struct pdc_dev *pdc, __u8 ep);
static void rx_command(struct pdc_dev *pdc, __u8 setup);
static void tx_command(struct pdc_dev *pdc);


/*--------------------------------------------------------------*
 *               initialization function declerations
 *--------------------------------------------------------------*/
static int __init pdc_module_init (void);
static void __exit pdc_module_cleanup (void);

static int      pdc_probe (struct isp1761_dev   *dev, isp1761_id *id);

static void pdc_remove (struct isp1761_dev      *dev);
static void pdc_init(struct isp1761_dev *dev);
static void pdc_connect(void);

/*--------------------------------------------------------------*
 *               local variable definitions
 *--------------------------------------------------------------*/

struct  pdc_dev                         usb_devices[1];         /* local interface */
static  struct isp1761_dev      *isp1761_dc_dev;        /* HOSAL interface */

static  spinlock_t                      pdc_rdwr_lock = SPIN_LOCK_UNLOCKED;
struct  pdc_pipe                        pdc_eps[PDC_MAX_PIPES];
struct  pdc_ep_desc                     pdc_ctrl_ep_desc[2];            /* Control IN and Control OUT */
u8      HS_Dev_flag;
EXPORT_SYMBOL(HS_Dev_flag);

extern struct pdc_bus                   *bus_dev;

#ifdef CONFIG_USB_OTG
int nctrl_msg = 0;
int g_pdc_dev_suspend=0;
#endif

/*--------------------------------------------------------------*
 *          Device Controller register access functions
 *--------------------------------------------------------------*/
/* static __inline__ */__u16    pdc_read16(__u16 reg)
{
    __u16       data=0;

    data = isp1761_reg_read16(isp1761_dc_dev, reg, data);

    return data;
}

/*static __inline__ */ void pdc_write16(__u16   reg, __u16      data)
{
    isp1761_reg_write16(isp1761_dc_dev, reg, data);
}

static __inline__ __u32 pdc_read32(__u16        reg)
{
    __u32   data =0;

    data = isp1761_reg_read32(isp1761_dc_dev, reg, data);

    return data;
}

static __inline__ void pdc_write32(__u32 reg, __u32     data)
{
    isp1761_reg_write32(isp1761_dc_dev, reg, data);

    return;
}


/*--------------------------------------------------------------*
 *          Device Controller endpoint related functions
 *--------------------------------------------------------------*/

/* Check to see if endpoint is full */
static __inline__ int checkendpoint(int endpoint)
{
    pdc_write32(ENDPT_INDEX_REGISTER, endpoint);
    return(pdc_read16(BUFFER_STATUS_REG) );

}



/* Read a packet out of a FIFO */
static int readendpoint(int endpoint, u8 *buffer, int length)
{
    int a,c;
    u32 *data_buf;

    data_buf = (u32*)buffer;

    /* Any data? */
    pdc_write32(ENDPT_INDEX_REGISTER, endpoint);

    if (!checkendpoint(endpoint)) return(0);
    /* read buffer length */
    c = pdc_read16(BUFFER_LGTH_REGISTER);

    dev_info("Length = %2x, endpt = %2x \n", length, endpoint );

    /* Read packet and align the bytes*/
    if (c>length) c=length;
    a=c;

	/* Bug fix for zero length packets */
	if (!a)
    {    
		pdc_write32(ENDPT_INDEX_REGISTER, endpoint);
		pdc_write32(CTRLFN_REGISTER, CNTRLFN_CLBUF);		
		c=a; 	
    }	
    else
    {
    	while(a>0) {
        	u32 w=pdc_read32(DATAPORT_REGISTER);
	        *data_buf=w;
    	    data_buf++;
        	a-=4;
	    }
    }

    /* Return bytes read */
    return(c);
}



/* Write a packet into the fifo */
static void writeendpoint(int endpoint, u8 *buffer, int length)
{
    int a;
    u32 *data_buf;
    data_buf = (u32*)buffer;

    dev_info("EP0x%08X writing packet length %d\n",endpoint,((unsigned short)length));

    /* write endpoint index */
//    pdc_write32(DMA_EP_INDX_REGISTER, 5);
    pdc_write32(ENDPT_INDEX_REGISTER, endpoint);        

    if (endpoint == EP0TX)
        pdc_write32(CTRLFN_REGISTER, CNTRLFN_DSEN);

    /* write buffer length to validate the buffer */
    pdc_write16(BUFFER_LGTH_REGISTER, length);

    /* Write the data  to  FIFO*/
    a=length;


    while(a>0) {
        pdc_write32(DATAPORT_REGISTER, (*data_buf) );
        data_buf++;
        a-=4;
    }
    udelay(5);  /*TODO fine tune */
}



/* 
 * Configure the Endpoints in the Hardware for those
 * pipes that are configured. For others Just diables
 * them. This function also adjusts the Endpoint max packet size to the
 * neareds available max packet size of the hardware.
 */
void pdc_configure_eps(struct pdc_dev   *pdc)
{
    __u16       ep_config = 0;
    __u8        handle;
    struct pdc_pipe     *pipe;
    __u16       mps=0;
    __u32       int_en;
    __u16   ep_cnt;

    dev_print("++ pdc_configure_eps(pdc=%p)\n",pdc);

    ep_cnt = 0;

    int_en = pdc_read32(INT_ENABLE_REGISTER);

    /* Configure all the Pipes (endpoint registers) in the device */
    for(handle=0;handle<PDC_MAX_PIPES;handle++) {

        pipe = pdc->ep_pipes+handle;            /* Get the pipe data structure */

        ep_config = 0;
        if(pipe->ep_state != PDC_PIPE_UNCONFIG) {
            /* pipe is configured */
            /* MPS, DBL BUFF, DIR, enable,disable */
            /* Use double buffering for non control pipes */

            if(pipe->ep_desc->ep_dir) ep_config |= EPCONFIG_EPDIR;
            if(handle > EP0TX) ep_config |= EP_TYP_DBLBUF;      

            if(pipe->ep_desc->attributes == PDC_EP_ISOCHRONOUS)
                ep_config |= PDC_EP_ISOCHRONOUS;
            else if(pipe->ep_desc->attributes == PDC_EP_INTERRUPT)
                ep_config |= PDC_EP_INTERRUPT;
            else
                ep_config |= PDC_EP_BULK;

            /* 
             * Set the maximum pkt size and get the device configuration
             * maximum packet size 
             */
            if(pipe->ep_desc->attributes != PDC_EP_ISOCHRONOUS) {
                if(HS_Dev_flag == FULL_SPEED)
                    mps = 64;
                else
                    mps = 512;

            } else {
                mps = 64;
            }

            int_en |= (IE_EP0RX << handle);
        }

        /* Configure the end point register */
        if(handle > EP0TX)
        {
            ep_cnt = handle;
            pdc_write32(ENDPT_INDEX_REGISTER,handle);
            pdc_write32(MAXP_SIZE_REGISTER,mps);
            pdc_write32(EP_TYP_REGISTER, ep_config); 
        }

    }

    for(handle=2; handle<=ep_cnt; handle++)
    {
        pdc_write32(ENDPT_INDEX_REGISTER,handle);                       
        ep_config = pdc_read32(EP_TYP_REGISTER);
        ep_config |= EP_TYP_ENABLE;
        pdc_write32(EP_TYP_REGISTER, ep_config);
    }

    pdc_write32(INT_ENABLE_REGISTER,int_en);

} /* End of pdc_configure_eps() */

void pdc_dev_control(unsigned long      opr) 
{

    __u32               data;
    __u16               data_u16;
    struct pdc_dev              *pdc = usb_devices;

    switch(opr) {

        case PDC_ENABLE:
#ifdef CONFIG_USB_OTG
            nctrl_msg = 0; /* Reset it */
#endif
            dev_info("pdc_dev_control: (PDC_ENABLE)\n");
            /* read Interrupt source register */
            data = pdc_read32(INTERRUPT_REGISTER);

            /* Bus is already in suspend state before enabling the DC
             * If the interrupt is already cleared, then we never get any
             * suspend interrupt later. So inform upper layers that the 
             * bus is suspended */
            if((data & INT_REG_SUSP) && pdc->pdc_bus && pdc->pdc_bus->notif )
                pdc->pdc_bus->notif(pdc->pdc_bus->context, PDC_BUSTATUS);


            /* Set device address & enable */
            pdc_write32(ADDRESS_REGISTER,ADDRESS_DEVEN|0);

            /* Set Interrupt config register 
               Default is set to 1us (0x1E). minimum pulse is 33.3ns */
            pdc_write32(INT_PULSE_SIZE_REG, DEFAULT_INT_PULSE_SIZE);
            pdc_write32(INT_CONFIG_REGISTER, DEFAULT_INT_CONFIG);

            /* Enable interrupts */
            pdc_write32(INT_ENABLE_REGISTER, IE_BASIC_CTRL_EP_MASK);

            /* Connect to the bus */
            pdc_write32(MODE_REGISTER, MODE_WKUPCS|MODE_CLKAON|MODE_GLINTENA); 
            break;

        case PDC_DISABLE:
            dev_info("pdc_dev_control: (PDC_DISABLE)\n");
            /* Go off bus */
            pdc_write32(ADDRESS_REGISTER,0);

            /* Turn off IRQs */
            pdc_write32(INT_ENABLE_REGISTER,0);

            /* Global IRQ disable & turn off softconnect */
#if 0
            data_u16 = pdc_read16(MODE_REGISTER);
            data_u16 &= (~MODE_INTENA);
            pdc_write16(MODE_REGISTER,data_u16);
#else
            pdc_write32(MODE_REGISTER,0);
#endif

            break;

        case PDC_CONNECT:
            dev_info("pdc_dev_control: (PDC_CONNECT)\n");
            /* check softconn bit defn */
            /* Connect it to the USB bus */
            data_u16 = pdc_read32(MODE_REGISTER);
            data_u16 |= MODE_SOFTCT;
            pdc_write32(MODE_REGISTER, data_u16);

            break;

        case PDC_DISCONNECT:
            dev_info("pdc_dev_control: (PDC_DISCONNECT)\n");
            /* Disconnect it from the USB bus */
            data_u16 = pdc_read32(MODE_REGISTER);
            data_u16 &= (~MODE_SOFTCT);
            pdc_write32(MODE_REGISTER, data_u16);

            break;
    }

    return;
} /* End of pdc_dev_control()  */



/* 
 * Deal with the received data on the USB pipe
 *  @pdc: pdc device data structure
 * @handle: USB pipe handle
 */
void rx_data(struct pdc_dev *pdc, pdc_pipe_handle_t     handle)
{
    struct pdc_pipe     *pipe = pdc->ep_pipes + handle;
    struct pdc_urb      *urb = pipe->urb;
    __u16                       status;
    __u8                        fifos = 0;
    __u32                       bytes, rcvd_bytes;

    u32 intr_src;
    u8 ep_stall;

    dev_print("++ rx_data(pdc=%p,handle=%x)\n",pdc,handle);

    if(in_interrupt()) {
        intr_src =pdc_read32(INTERRUPT_REGISTER);
        intr_src &= IE_EP1RX;
        if(intr_src)
            pdc_write32(INTERRUPT_REGISTER, intr_src);

        pdc_write32(ENDPT_INDEX_REGISTER, handle);
        status = pdc_read16(BUFFER_STATUS_REG);
        pipe->ep_status = status;
    } else {
        status = pipe->ep_status;
    }


    if (status == BUF1_AND_2_FILLED)
        fifos = 2;
    else
        fifos = 1;

    pdc_write32(ENDPT_INDEX_REGISTER, handle);
    ep_stall = (pdc_read32(CTRLFN_REGISTER) & CNTRLFN_STALL);

    if (!(status) || ep_stall) {
        return ;
    }


    while((fifos--) && urb) {

        /* How many bytes do we need? */
        bytes = urb->transfer_buffer_length - urb->actual_length;

        if(bytes > pipe->ep_desc->max_pkt_size) bytes = pipe->ep_desc->max_pkt_size;

        rcvd_bytes = readendpoint(handle,&(((u8*)urb->transfer_buffer)[urb->actual_length]),bytes);

        urb->actual_length += rcvd_bytes;

        /* Clear the full bits */
        if(pipe->ep_status & BUF0_FILLED) {
            pipe->ep_status &= ~(BUF0_FILLED);
        } else if(pipe->ep_status & BUF1_FILLED) {
            pipe->ep_status &= ~(BUF1_FILLED);
        }

        /* Did we receive a short packet?? */
        if(rcvd_bytes < bytes) {
            urb->status = PDC_SHORT_PACKET;
            dev_check("Short packet\n");
        }


        /* Complete the urb if all the bytes are received from host or terminated
         * because of short packet
         */
        if(urb->transfer_buffer_length == urb->actual_length|| urb->status == PDC_SHORT_PACKET) {
            pipe->urb = urb->next;
            if(urb->status == PDC_URB_PENDING) {
                urb->status = PDC_URB_COMPLETE;

            }

            if(urb->complete) urb->complete(urb);
            urb = pipe->urb;
        }

    } //while(fifo--)
} /* End of rx_data() */




/* 
 * Deal with the transmit data on a USB pipe
 * @pdc : pdc device data structure
 * @pipe :  USB pipe
 */
void tx_data(struct pdc_dev *pdc, pdc_pipe_handle_t     handle, int kick)
{
    int txstat,tofill=1;
    struct pdc_pipe *pipe = pdc->ep_pipes + handle;
    struct pdc_urb      *ep_urb = pipe->urb;
    int                         usb_txsize;
    u32 intr_src;

    dev_print("++ tx_data(handle=%x, kick=%x)\n", handle, kick);

    if (kick == 2) {



        /* BUG: sending NULL data packet after stall is not valid .....*/
        /* Send NULL packet if DATA PID is not cleared by the HW 
         * This case assumed that tx_data(2) is used only for endpoint EP2 */

        intr_src =pdc_read32(INTERRUPT_REGISTER);
        intr_src &= IE_EP1TX;
        if(intr_src)
            pdc_write32(INTERRUPT_REGISTER, intr_src);

        pipe->txrx_idle = 1;


        if(ep_urb) {

            ep_urb->actual_length = 0;
            if(ep_urb->next)    ep_urb->next->actual_length = 0;
            kick = 1;
        }
        else return;
    }

    if (!kick) {
        /* Get status/clear IRQ */
        intr_src =pdc_read32(INTERRUPT_REGISTER);
        intr_src &= IE_EP1TX;   /*IE_EP0RX << handle */
        if(intr_src)
            pdc_write32(INTERRUPT_REGISTER, intr_src);

        pdc_write32(ENDPT_INDEX_REGISTER, handle);
        txstat = pdc_read16(BUFFER_STATUS_REG);

        /* Successfully sent some stuff: bump counts & reset buffer */
        ep_urb = pipe->urb;

        /* If last packet was short, and there's nothing in the buffer
           to send, then just stop here with TX disabled */
        /* XXX: when the data transfer is over, go to silent mode */
        if(ep_urb && (ep_urb->actual_length == ep_urb->transfer_buffer_length) && !(txstat)) {
            /* URB Complete */
            pipe->urb = ep_urb->next;
            ep_urb->actual_length = ep_urb->transfer_buffer_length;
            ep_urb->status = PDC_URB_COMPLETE;
            if(ep_urb->complete)        ep_urb->complete(ep_urb);
            ep_urb = pipe->urb;
        }

        if(!ep_urb) {
            pipe->txrx_idle = 1;
            return;
        }
    }

    ep_urb = pipe->urb;

    /* Work out how many FIFOs we can fill */
    tofill = 2;
    intr_src =pdc_read32(INTERRUPT_REGISTER);
    intr_src &= IE_EP1TX;
    if(intr_src)
        pdc_write32(INTERRUPT_REGISTER, intr_src);

    pdc_write32(ENDPT_INDEX_REGISTER, handle);
    txstat = pdc_read16(BUFFER_STATUS_REG);
    if(txstat & BUF0_FILLED) tofill--;
    if(txstat & BUF1_FILLED) tofill--;

    dev_check(" tx stat= %x \n", txstat);

    /* Bug fix for zero length packets */
    if (!ep_urb->transfer_buffer_length && ((__u8*)ep_urb->transfer_buffer) == NULL)
	  {	   
		  pdc_write32(ENDPT_INDEX_REGISTER, handle);
		  pdc_write32(DATAPORT_REGISTER, 0);
		  pdc_write16(BUFFER_LGTH_REGISTER, 0);
    }	
	  else
	  {
	    /* The problem with this for BOT protocol is in the IN data stage
	     * after sending the data, host expects status (IN) data and we are
    	 * sending a NULL packet which is causing whole device to re-enumerate
	     * In this case the data is multiple of MaxPacketSize 
	     * So only if tx_used is there then send the data */
	     
    	/* While we can send stuff... (this will fill both FIFOs) */        
	    if((kick==1)&&ep_urb->picdata)   // Call from pdc_submit_urb -> start pic transfer
      {
//        printk("Start Pic Transfer!\n");
        pdc_write32(0x40000,ep_urb->buffer_start_address+ep_urb->actual_length);      // Neu 0x40000 und 0x40004  // Alt 0x800 und 0x804
//        tofill=1;
      }
	    if(pipe->ep_desc->max_pkt_size==64)   //!!!DK: Bei USB1.1 kein Doublebuffering, funktioniert irgendwie nicht!
	      tofill=1;
	    
	    if(ep_urb->picdata)
      {
        usb_txsize = ep_urb->transfer_buffer_length - ep_urb->actual_length;

        pdc_write32(ENDPT_INDEX_REGISTER, handle);        
        switch(tofill)
        {
          case 2:
            if(usb_txsize >= 2 * pipe->ep_desc->max_pkt_size)
              usb_txsize = 2 * pipe->ep_desc->max_pkt_size;
            else if((usb_txsize >= 1 * pipe->ep_desc->max_pkt_size) && (usb_txsize < 2 * pipe->ep_desc->max_pkt_size))
              usb_txsize = 1 * pipe->ep_desc->max_pkt_size;
            else
              pdc_write16(BUFFER_LGTH_REGISTER, usb_txsize);
            break;      
          case 1:
            if(usb_txsize >= 1 * pipe->ep_desc->max_pkt_size)
              usb_txsize = 1 * pipe->ep_desc->max_pkt_size;
            else
              pdc_write16(BUFFER_LGTH_REGISTER, usb_txsize);
            break;
          case 0:
            printk("No free buffer!\n");
            usb_txsize = 0;
            return;
//            break;
        }
//        printk("tf:%d sz:%d ",tofill,usb_txsize);
        if(usb_txsize==2 * pipe->ep_desc->max_pkt_size)
        {
          pdc_write32(0x40004,usb_txsize/8);
          udelay(1);
          pdc_write32(0x40004,usb_txsize/8);
        }
        else
          pdc_write32(0x40004,usb_txsize/4);
	      ep_urb->actual_length += usb_txsize;
    	  /* Not idle anymore */
	      pipe->txrx_idle=0;
      }
      else
      {
	      while( (ep_urb) && (tofill>0) && 
    	          ep_urb->actual_length != ep_urb->transfer_buffer_length) 
	      {
    	      /* Fill local packet buffer from TX buffer: if there's nothing
        	     to send (there might be: we need to be able to send zero
	             length packets to terminate a transfer of an exact multiple
	             of the buffer size), then we'll be sending a 0 byte
	             packet */    

    	      usb_txsize = ep_urb->transfer_buffer_length - ep_urb->actual_length;

        	  if ((usb_txsize)> pipe->ep_desc->max_pkt_size) usb_txsize=pipe->ep_desc->max_pkt_size;
          
            writeendpoint(handle, (((__u8*)ep_urb->transfer_buffer)+ep_urb->actual_length), usb_txsize);

	          ep_urb->actual_length += usb_txsize;

    	      /* Not idle anymore */
	          pipe->txrx_idle=0;

	          /* Filled another buffer */
	          tofill--;
    	  }
      }
	}

} /* End of tx_data() */

/* Deal with incoming packet on the control endpoint */
void rx_command(struct pdc_dev *pdc, __u8 setup)
{
    /* Get receiver status */    
    u8 command[8];
    unsigned long       bytes, rcvd_bytes;
    u32 intr_src;
    struct pdc_pipe *pipe;
    struct pdc_urb  *urb;

    /* Is this a setup packet? */  
    if(setup) {
        int i;
        /* clear Setup int if not cleared in INTR routine */

        intr_src =pdc_read32(INTERRUPT_REGISTER);
        intr_src &= IE_EP0SETUP;
        if(intr_src)
            pdc_write32(INTERRUPT_REGISTER, intr_src);

        /* Read the packet into our buffer */
        if ((i=readendpoint(EP0SETUP,command,sizeof(command)))!=sizeof(command)) {

            pdc_write32(ENDPT_INDEX_REGISTER, EP0TX);
            pdc_write32(CTRLFN_REGISTER, CNTRLFN_STALL);

            pdc_write32(ENDPT_INDEX_REGISTER, EP0RX);
            pdc_write32(CTRLFN_REGISTER, CNTRLFN_STALL);

            dev_check("pdc:Short USB control read (%d bytes)!\n",i);
            return;
        }

        /* Acknowledge out endpoints, then clear; the
           arrival of the setup packet disables the clear command
           which is issued in readendpoint() */

        /* Clear all the previous data transfers on the IN endpoint */
        pipe = pdc->ep_pipes + EP0RX;
        urb = pipe->urb;

        /* For each URB complete the URB */
        while(urb) {

            printk(" PDC_SETUP_OVERRITE \n");
            urb->status = PDC_SETUP_OVERRITE;
            pipe->urb = urb->next;
            if(urb->complete)   urb->complete(urb);
            urb = pipe->urb;
        }


        /* Call the notification function of Configuration driver */
        pipe = pdc->ep_pipes + EP0RX;

        if((pipe->pipe_desc) && (pipe->pipe_desc->notify))      {
            pipe->pipe_desc->notify(PDC_SETUP_COMMAND,
                    pipe->pipe_desc->priv,
                    command);
        }


    } else {
        /* If not a setup packet, it must be an OUT packet */ 

        intr_src =pdc_read32(INTERRUPT_REGISTER);
        intr_src &= IE_EP0RX;
        if(intr_src)
            pdc_write32(INTERRUPT_REGISTER, intr_src);

        pipe = pdc->ep_pipes + EP0RX ;

        urb = pipe->urb;

        if(urb) {
            /* How many bytes do we need? */
            bytes = urb->transfer_buffer_length - urb->actual_length;

            rcvd_bytes = readendpoint(EP0RX,&(((__u8*)urb->transfer_buffer)[urb->actual_length]),bytes);

            urb->actual_length += rcvd_bytes;

            /* Did we receive a short packet?? */
            if(rcvd_bytes < bytes) urb->status = PDC_SHORT_PACKET;

            /* Complete the urb if all the bytes are received from host or terminated
             * because of short packet
             */
            if(urb->transfer_buffer_length == urb->actual_length|| urb->status == PDC_SHORT_PACKET) {
                /* change for Compliance on 05-08-05 start.
                   This change is needed  for handling the Control OUT request */
                if(urb->transfer_buffer_length)
                {
                    pdc_write16(ENDPT_INDEX_REGISTER, EP0TX);
                    pdc_write16(CTRLFN_REGISTER, CNTRLFN_STATUS_EM_ACK);
                }       
                /*change ends */
                pipe->urb = urb->next;
                if(urb->status == PDC_URB_PENDING) {
                    urb->status = PDC_URB_COMPLETE;
                }
                if(urb->complete) urb->complete(urb);
                urb = pipe->urb;
            }
        } else {
            /* Clear the buffer, Do we need to wait for the appl?? let us see */
            pdc_write32(ENDPT_INDEX_REGISTER,EP0RX);
            pdc_write32(CTRLFN_REGISTER,CNTRLFN_CLBUF);

        }

    }
} /* End of rx_command */

/* TX command */
void tx_command(struct pdc_dev *pdc)
{
    struct pdc_pipe             *pipe = pdc->ep_pipes + EP0TX;

    struct pdc_urb              *pipe_urb;
    __u16                               len;
    u32 intr_src;
    u16 icnt;

    dev_print("++ tx_command(void)\n"); 
    /* Read last status & discard */

    intr_src =pdc_read32(INTERRUPT_REGISTER);
    intr_src &= IE_EP0TX;
    if(intr_src)
        pdc_write32(INTERRUPT_REGISTER, intr_src);

    pipe_urb = pipe->urb;

    if(pipe_urb) {

        len = pipe_urb->transfer_buffer_length - pipe_urb->actual_length;

        if(len > pipe->ep_desc->max_pkt_size)   len = pipe->ep_desc->max_pkt_size;

        pdc_write32(ENDPT_INDEX_REGISTER,0x01);
        icnt = pdc_read32(CTRLFN_REGISTER);

        writeendpoint(EP0TX, &(((u8*)pipe_urb->transfer_buffer)[pipe_urb->actual_length]), len);

        pipe_urb->actual_length += len;

        if(pipe_urb->actual_length == pipe_urb->transfer_buffer_length) {

            if(pipe_urb->transfer_buffer_length)
            {
                pdc_write32(ENDPT_INDEX_REGISTER, EP0RX);
                pdc_write32(CTRLFN_REGISTER, CNTRLFN_STATUS_EM_ACK);

            }   
            pipe_urb->status = PDC_URB_COMPLETE;
            pipe->urb = pipe_urb->next;
            if(pipe_urb->complete) pipe_urb->complete(pipe_urb);
        }

    }

    if(!pipe->urb) pipe->txrx_idle = 1;

} /* End of tx_command */

#undef DEBUG_IRQ

/* Interrupt Service Routine */
void pdc_isr(struct isp1761_dev *dev, void *isr_data)
{

    struct pdc_dev                      *pdc = (struct pdc_dev*)isr_data;
    unsigned int                        evnt, non_ctrl_ep_event, ep_event;
    struct pdc_config_dev       *bus = pdc->pdc_bus;
    pdc_pipe_handle_t           handle;
    struct pdc_pipe                     *pipe;
    struct pdc_ep_desc  *ep_desc;
    __u32 tmpdata1, tmpdata2;
    u32 intr_src;
    u32 temp_data;

    while(1) {
        /* Read IRQ status, masking with possible valid IRQs */

//        pio_debug_write(3);
        evnt= dev->int_reg& (IE_EP0SETUP|IE_EP0RX|IE_EP0TX|IE_NON_CTRL_EP_MASK|IE_RST|IE_SUSP|IE_VBUS|IE_HS_STA|IE_RESM|IE_DMA);

        temp_data = pdc_read32(INT_ENABLE_REGISTER);
        evnt = evnt & temp_data;

        /* No irq? */
        if (evnt==0) {
            return;
        }
#if 1
        non_ctrl_ep_event = evnt & IE_NON_CTRL_EP_MASK;

        handle = EP2RX;

        while(non_ctrl_ep_event) {

//        pio_debug_write(7);
            ep_event = IE_EP0RX << handle;

            if( non_ctrl_ep_event & ep_event) {

                pipe = pdc->ep_pipes + handle;

                if(pipe->ep_desc && pipe->pipe_desc) {
                    if(pipe->ep_desc->ep_dir) {
                        /* In pipe */
#ifdef DEBUG_IRQ
printk("IN_Pipe\n");
#endif
//        pio_debug_write(15);
                        tx_data(pdc, handle, 0);        /* Data trasmission on In pipe */
//        pio_debug_write(7);

                    } else {
                        /* Out pipe */
#ifdef DEBUG_IRQ
printk("OUT_Pipe\n");
#endif
                        rx_data(pdc, handle);           /* Data reception on OUT pipe */

                    }

                } else {
                    /* What to do?? received interrupt on an unconfigured, unopen pipe */
                    dev_info(" int on unconfigured pipe \n");
                }

                non_ctrl_ep_event &= ~ep_event;

            }

            handle++;
//        pio_debug_write(3);
        }

#endif
#if 0        
        
        if(evnt&IE_EP1RX)
        {
          rx_data(pdc, 2);           /* Data reception on OUT pipe */
        }
        if(evnt&IE_EP1TX)
        {
          tx_data(pdc, 3, 0);        /* Data trasmission on In pipe */
        }
        if(evnt&IE_EP2RX)
        {
          rx_data(pdc, 4);           /* Data reception on OUT pipe */
        }
        if(evnt&IE_EP2TX)
        {
//        pio_debug_write(7);
          tx_data(pdc, 5, 0);        /* Data trasmission on In pipe */
//        pio_debug_write(3);
        }
#endif
        
        if (evnt&IE_EP0SETUP) {
#ifdef DEBUG_IRQ
printk("EP0STP\n");
#endif
            rx_command(pdc,1);  /* Control SETUP */
#ifdef CONFIG_USB_OTG
            nctrl_msg++;
#endif
        }
        if (evnt&IE_EP0RX) {

#ifdef DEBUG_IRQ
printk("EP0RX\n");
#endif
            rx_command(pdc,0);  /* Control Out */
        }

        if (evnt&IE_EP0TX) {

#ifdef DEBUG_IRQ
printk("EP0TX\n");
#endif
            tx_command(pdc);    /* COntrol IN */
        }


        if (evnt&IE_RST) {

#ifdef DEBUG_IRQ
printk("RST\n");
#endif
            /* Inform the configuration driver 
               Initialize the controller and connect to the bus */

            if(bus && bus->notif)       bus->notif(bus->context, PDC_RESET);
            HS_Dev_flag = FULL_SPEED;
            pdc_init(dev); 
            pdc_connect();

            /* Re-initialise endpoint for FULL SPEED  */                    
            ep_desc = pdc->pdc_bus->ep_desc; 
            ep_desc++;
            ep_desc->max_pkt_size = 64; /* Maximum packet size */
            ep_desc++;
            ep_desc->max_pkt_size = 64; /* Maximum packet size */
        }

        if (evnt&IE_SUSP) 
        {

#ifdef DEBUG_IRQ
printk("SUSP\n");
#endif
            if(bus && bus->notif)
                bus->notif(bus->context, PDC_SUSPEND);
            intr_src =pdc_read32(INTERRUPT_REGISTER);
            intr_src &= IE_SUSP;
            if(intr_src)
                pdc_write32(INTERRUPT_REGISTER, intr_src);

            tmpdata1 = pdc_read32(MODE_REGISTER);
            tmpdata2  = tmpdata1|MODE_GOSUSP;
            pdc_write32(MODE_REGISTER, tmpdata2);
            pdc_write32(MODE_REGISTER, tmpdata1);
        }

        if (evnt&IE_RESM) {
            /* Inform the configuration driver  */
            if(bus && bus->notif)       bus->notif(bus->context, PDC_RESUME);
        }


        if (evnt&IE_VBUS) {
            printk("USB Device connected\n");
            intr_src =pdc_read32(INTERRUPT_REGISTER);
            intr_src &= IE_VBUS;
            if(intr_src)
                pdc_write32(INTERRUPT_REGISTER, intr_src);


        }
        if (evnt&IE_HS_STA) {
            /* Reconfigure the device endpoints if the device is in FS */
            printk("Device High Speed Mode \n");
            if(HS_Dev_flag == FULL_SPEED)
            {
                HS_Dev_flag = HIGH_SPEED;
                /* Re-initialise endpoint for high speed */

                ep_desc = pdc->pdc_bus->ep_desc; 
                ep_desc++;
                ep_desc->max_pkt_size = 512; /* Maximum packet size */
                ep_desc++;
                ep_desc->max_pkt_size = 512; /* Maximum packet size */

            }
        }
        if (evnt&IE_DMA) {

            /* Handle DMA Int source register  */
            /* TODO */
        }       


        pdc_write32(UNLOCK_DEVICE_REG, 0xaa37);

        dev->int_reg =pdc_read32(INTERRUPT_REGISTER);
        if(dev->int_reg) pdc_write32(INTERRUPT_REGISTER, dev->int_reg);
    }
//        pio_debug_write(1);

} /* End of pdc_isr */





/*------------------------------------------------------------------------*
 *                  External Interface functions                          *
 *------------------------------------------------------------------------*/

/*
 * Get the current frame number on the USB bus
 * @cd  class driver identifier
 */
unsigned long   pdc_get_frame_number(struct pdc_class_drv *cd)
{
    unsigned long flags;
    unsigned long frame_number;


    dev_print("++ pdc_get_frame_number(cd=%p)\n",cd);
    spin_lock_irqsave(&pdc_rdwr_lock, flags);
    frame_number = pdc_read32(FRAME_NUMBER_REGISTER);
    frame_number &= 0x7FF; 
    spin_unlock_irqrestore(&pdc_rdwr_lock, flags);

    return frame_number;
} /* End of pdc_get_frame_number() */


/* 
 * Set the address of the device
 * @context is the device context
 * @address is the device address on the bus
 */
int pdc_set_device_address(unsigned long context,unsigned short address)
{
    dev_info("pdc_set_device_address(context=%x,address=%x\n",context, address);// 999  
    pdc_write32(ADDRESS_REGISTER, (ADDRESS_DEVEN|address));
    return 0;
} /* End of pdc_set_device_address() */

/*
 * Configure the device endpoints and Notification function
 */
int pdc_configure_device(struct pdc_config_dev *config)
{
    struct pdc_ep_desc  *ep_desc = config->ep_desc;
    struct pdc_pipe             *pipe;
    struct pdc_dev              *pdc = usb_devices;
    pdc_pipe_handle_t   handle;
    __u8                                index;


    dev_print("++ pdc_configure_device(config=%p)\n",config);

    /* Initialize all pipe varaiables except the control ones */
    for(handle=2; handle < PDC_MAX_PIPES; handle++) {
        pipe = pdc->ep_pipes + handle;
        pipe->ep_desc = NULL;
        pipe->pipe_desc = NULL;
        pipe->ep_state = PDC_PIPE_UNCONFIG;
    }

    /* Go through all the end points in the list */
    for(index = 0; index < config->num_eps; index++) {

        if(ep_desc->ep_num != 0 && ep_desc->ep_num < PDC_MAX_EPS) {

            /* Get the pipe handle for the Endpoint */
            handle = pdc_usb_to_epreg((ep_desc->ep_num),(ep_desc->ep_dir));

            pipe = pdc->ep_pipes + handle;      /* Get the pipe descriptor */

            pipe->ep_desc = ep_desc;

            pipe->ep_state = PDC_PIPE_CONFIG;

        }

        ep_desc++;
    }


    /* Copy the configuration driver information */
    pdc->pdc_bus = config;

    /* Configure the endpoints in the Hardware */
    pdc_configure_eps(pdc);

    return 0;
} /* End of pdc_configure_device() */


/*
 * Open a data transfer pipe
 * @pipe_desc:  pipe information
 */
pdc_pipe_handle_t pdc_open_pipe(struct pdc_pipe_desc *pipe_desc)
{
    struct pdc_pipe             *pipe;
    pdc_pipe_handle_t   handle = PDC_INV_PIPE_HANDLE;
    struct pdc_dev              *pdc=usb_devices;

    dev_print("++ pdc_open_pipe(pipe_desc=%p)\n",pipe_desc);

    if(pipe_desc->ep < PDC_MAX_EPS) {
        handle = pdc_usb_to_epreg((pipe_desc->ep),(pipe_desc->ep_dir));

        pipe = pdc->ep_pipes + handle;  

        /* If pipe is configured before, then open the pipe */
        if(pipe->ep_state == PDC_PIPE_CONFIG) {
            /* Only configured pipes can be opened */
            pipe->ep_state = PDC_PIPE_OPEN;
            pipe->pipe_desc = pipe_desc;
            pipe->txrx_idle = 1;
            if(pipe_desc->ep_dir==PDC_EP_DIR_IN)
            {
//		          printk("handle:0x%x\n",handle);
		          pdc_write32(ENDPT_INDEX_REGISTER, handle);
		          pdc_write32(DMA_EP_INDX_REGISTER, 9);
//		          pdc_write32(DMA_CMD_REGISTER, 0x0f);
		          pdc_write32(CTRLFN_REGISTER, CNTRLFN_CLBUF);		
              udelay(10);
		          pdc_write32(ENDPT_INDEX_REGISTER, handle);
		          pdc_write32(DMA_EP_INDX_REGISTER, 9);
//		          pdc_write32(DMA_CMD_REGISTER, 0x0f);
		          pdc_write32(CTRLFN_REGISTER, CNTRLFN_CLBUF);		
            }
        } else {
            /* else return invalid handle */
            handle = PDC_INV_PIPE_HANDLE;
        }

        if(pipe_desc->ep == 0) {
            /* 
             * This is a control end point to the user and control out pipe for us,
             * so open a control in pipe also
             */

            pipe = pdc->ep_pipes + handle + 1;

            pipe->ep_state = PDC_PIPE_OPEN;
            pipe->pipe_desc = pipe_desc;
        }
    }


    return handle;
} /* End of pdc_open_pipe() */

EXPORT_SYMBOL(pdc_open_pipe);

/*
 * Close an already opened pipe
 * @pipe_handle:        pipe handle
 */
void pdc_close_pipe(pdc_pipe_handle_t   pipe_handle)
{
    __u8        ep;
    struct pdc_pipe *pipe;
    struct pdc_dev              *pdc=usb_devices;
    u32 ep_type;

    dev_print("++ pdc_close_pipe(pipe_handle=%x)\n",pipe_handle);

    if(pipe_handle < PDC_MAX_PIPES) {

        pipe = pdc->ep_pipes + pipe_handle;

        /* If the pipe is opened then cleanup the stuff */
        if(pipe->ep_state == PDC_PIPE_OPEN) {
            dev_info(" pdc_close_pipe , pipe = %02x, epdes_epno = %02x, epdes_epdir= %02x,pipdes_ep= %02x \n", 
                    pipe->pipe, pipe->ep_desc->ep_num,pipe->ep_desc->ep_dir,pipe->pipe_desc->ep);
            
            if(pipe->pipe_desc->ep_dir==PDC_EP_DIR_IN)
            {
//              printk("pipe_handle:%d\n",pipe_handle);
              pdc_write32(DMA_EP_INDX_REGISTER, 9);  
              pdc_write32(ENDPT_INDEX_REGISTER, pipe_handle);  
              pdc_write32(CTRLFN_REGISTER,CNTRLFN_CLBUF);

              pdc_write32(DMA_EP_INDX_REGISTER, 9);  
              pdc_write32(ENDPT_INDEX_REGISTER, pipe_handle);  
              pdc_write32(CTRLFN_REGISTER,CNTRLFN_CLBUF);

              ep_type = pdc_read32(EP_TYP_REGISTER);     //Disable EP
              ep_type &= (~EP_TYP_ENABLE);
              pdc_write32(EP_TYP_REGISTER, ep_type);
            }
            pipe->pipe_desc =  NULL;
            pipe->ep_state = PDC_PIPE_CONFIG;

        }
//            if(pipe_desc->ep_dir==PDC_EP_DIR_IN)
        /* 
         * Get the endpoint and if it is control ep then close
         * the control IN pipe also
         */
        ep = pdc_pipe_to_epreg(pipe_handle); 

        if(ep == 0) {
            pdc_close_pipe(pipe_handle + 1);
        }
    }
} /* End of pdc_close_pipe() */

EXPORT_SYMBOL(pdc_close_pipe);

int pdc_submit_control_urb(struct pdc_dev       *pdc, struct pdc_urb *urb_req)
{
    struct pdc_pipe     *pipe;
    struct pdc_urb      *urb;
    unsigned long       length;
    pdc_pipe_handle_t   handle = urb_req->pipe;

    dev_print("++ pdc_submit_control_urb(urb=%p)\n",urb_req);

    if(urb_req->operation == PDC_OPR_WRITE) handle += 1;        /* Control IN pipe */

    pipe = pdc->ep_pipes + handle;

    urb = pipe->urb;

    urb_req->next = 0;
    if(urb != urb_req)
    {
        /* Insert the URB in to the pipe URB queue (FIFO) */
        if(!urb) {
            pipe->urb  = urb_req;
        } else {
            while(urb) {
                if(urb->next != NULL)   urb = urb->next;
                else urb->next = urb_req;
            }
        }
    }

    /* set the output parameters */
    urb_req->actual_length = 0;
    urb_req->next = NULL;

    if(urb_req->operation == PDC_OPR_WRITE) {

        if(pipe->urb == urb_req)        {
            length = urb_req->transfer_buffer_length;
            if(urb_req->transfer_buffer_length > pipe->ep_desc->max_pkt_size) {
                length = pipe->ep_desc->max_pkt_size;
            }

            if(length != 0)
            {
                writeendpoint(handle,((u8*)urb_req->transfer_buffer),length);                           
                urb_req->actual_length += length;
            }   

        }

        if(urb_req->actual_length == urb_req->transfer_buffer_length) {

            if(urb_req->transfer_buffer_length)
            {
                udelay(100); /*TODO fine tune wait for TX delay */
                return 0; 

            }
            else
            {
                udelay(100); /*TODO fine tune wait for TX delay */
                pdc_write32(ENDPT_INDEX_REGISTER,EP0TX);
                pdc_write32(CTRLFN_REGISTER,CNTRLFN_STATUS_EM_ACK);

            }
            dev_info(" actual= %x \n", urb_req->actual_length);
            urb_req->status = PDC_URB_COMPLETE;
            pipe->urb = urb_req->next;
            if(urb_req->complete) urb_req->complete(urb);
        }

    } else if(urb_req->operation == PDC_OPR_READ) {
        /* change for Compliance on 05-08-05 start. 
        This change is needed to handle the Control OUT request */
        if(pipe->urb == urb_req)        {

            pdc_write32(ENDPT_INDEX_REGISTER, EP0RX);   
            pdc_write32(CTRLFN_REGISTER, CNTRLFN_DSEN);
        }
    }
    	/* change ends */

    return 0;
}

/*
 * Submit an urb for data transfer
 * @urb: urb request block
 */
int pdc_submit_urb(struct pdc_urb *urb_req)
{
    struct pdc_pipe     *pipe;
    struct pdc_urb      *urb;
    unsigned long flags;
    struct pdc_dev              *pdc=usb_devices;

    dev_check("s_urb(urb=%p),l=%d\n",urb_req, urb_req->transfer_buffer_length);
    udelay(100);
    dev_print("++ pdc_submit_urb(urb=%p)\n",urb_req);

    if(urb_req->pipe > PDC_MAX_PIPES)   return -EINVAL;

    if(urb_req->pipe < 2)       return pdc_submit_control_urb(pdc, urb_req);

    pipe = pdc->ep_pipes + urb_req->pipe;
    dev_info(" pdc_ep_pipe = %x, urb_req_pipe = %x \n",  pdc->ep_pipes->pipe, urb_req->pipe);
    spin_lock_irqsave(&pdc_rdwr_lock, flags);

    urb = pipe->urb;

    /* Insert the URB in to the pipe URB queue (FIFO) */
    urb_req->next = 0;
    if(urb != urb_req)
    {
        if(!urb) {
            pipe->urb  = urb_req;
        } else {
            unsigned int i=0;

            while(urb->next != NULL) {
                if(i++<100) {
                    dev_check("urb = %p, urb_req = %p\n", urb, urb_req);
                }
                urb = urb->next;
            }
            urb->next = urb_req;

        }
    }
    else
    {
        printk("ERROR: urb = urb_req = %p\n", urb);
    }

    /* set the output parameters */
    urb_req->actual_length = 0;
    urb_req->next = NULL;
    dev_info(" urb_req->operation = %x \n", urb_req->operation);
    udelay(100);

    if(urb_req->operation == PDC_OPR_READ) {

        /* 
         * handle the read request, if there are no URB's,
         * call the rx_data function 
         */
        if(pipe->urb == urb_req)        {

            rx_data(pdc, urb_req->pipe);
        }

    } else if(urb_req->operation == PDC_OPR_WRITE) {

        /*
         * Handle the write request.
         * If the Pipe is not stalled and the transmission is idle
         * start the transmission
         */

        pdc_write32(ENDPT_INDEX_REGISTER, urb_req->pipe);
        pipe->ep_status = (pdc_read32(CTRLFN_REGISTER) & CNTRLFN_STALL) ? PDC_PIPE_STALL : PDC_PIPE_UNSTALL ;

        if(pipe->txrx_idle && pipe->ep_status != PDC_PIPE_STALL) {
            tx_data(pdc, urb_req->pipe, 1);     /* Kick the transfer */
        }

    }
    udelay(100);
    spin_unlock_irqrestore(&pdc_rdwr_lock, flags);

    return 0;
} /* End of pdc_submit_urb */

EXPORT_SYMBOL(pdc_submit_urb);

/*
 * Cancel an urb for data transfer
 * @urb: urb request block
 */
int pdc_cancel_urb(struct pdc_urb *urb_req)
{
    struct pdc_pipe     *pipe;
    struct pdc_urb      *urb, *prev_urb;
    unsigned long flags;
    struct pdc_dev              *pdc=usb_devices;

    dev_print("++ pdc_cancel_urb(%p)\n",urb_req);

    /* Check the pipe value */
    if(urb_req->pipe > PDC_MAX_PIPES)   return -EINVAL;

    pipe = pdc->ep_pipes + urb_req->pipe;

    if((urb_req->pipe == 0) && (urb_req->operation == PDC_OPR_WRITE)) {
        /* Special case of control IN pipe */
        pipe++;
    }

    spin_lock_irqsave(&pdc_rdwr_lock, flags);

    prev_urb = NULL;
    urb = pipe->urb;

    /* 
     * search for the URB , The pipe could be either IN or OUT so there is
     * only one URB list 
     */
    while((urb!=NULL) && (urb_req != urb)) {
        prev_urb = urb;
        urb = urb->next;
    }

    /* 
     * found URB, then Un link it from the list
     */
    if(urb) {
        urb->status = PDC_URB_COMPLETE;
        if(prev_urb)    prev_urb->next = urb->next;
        else pipe->urb = urb->next;
    }

    spin_unlock_irqrestore(&pdc_rdwr_lock, flags);

    return 0;
} /* End of pdc_cancel_urb */

EXPORT_SYMBOL(pdc_cancel_urb);

/*
 * Pipe operations
 * @pipe_opr:
 */
int pdc_pipe_operation(struct pdc_pipe_opr   *pipe_opr)
{
    struct pdc_pipe     *pipe;
    __u16       status;
    struct pdc_dev              *pdc=usb_devices;
    pdc_pipe_handle_t   handle = pipe_opr->handle;
    u32 ep_type;

    dev_print("++ pdc_pipe_control(pipe_opr=%p)\n",pipe_opr);

    dev_check("pdc_pipe_operation \n");

    if(handle < PDC_MAX_PIPES) {

        pipe = pdc->ep_pipes + handle;

        switch(pipe_opr->opr) {

            case PDC_PIPE_STALL:
                dev_check(" stall pipe stall \n");

                pdc_write32(ENDPT_INDEX_REGISTER, handle);
                pdc_write32(CTRLFN_REGISTER, CNTRLFN_STALL);

                if(handle == EP0RX) 
                {
                    /* Stall control in also */
                    pdc_write32(ENDPT_INDEX_REGISTER, handle +1);
                    pdc_write32(CTRLFN_REGISTER, CNTRLFN_STALL);

                }
                break;

            case PDC_PIPE_UNSTALL:
                dev_check("stall pipe unstall \n");
                pdc_write32(ENDPT_INDEX_REGISTER, handle);
                ep_type = pdc_read32(CTRLFN_REGISTER);
                pdc_write32(CTRLFN_REGISTER, 0);

                if(handle == EP0RX) 
                {
                    pdc_write32(ENDPT_INDEX_REGISTER, handle +1);  
                    pdc_write32(CTRLFN_REGISTER, 0);
                }

                if(handle > EP0TX) {
                    if(pipe->ep_desc->ep_dir & PDC_EP_DIR_IN)
                    {
                        pdc_write32(ENDPT_INDEX_REGISTER, handle);  
                        ep_type = pdc_read32(CTRLFN_REGISTER);
                        pdc_write32(CTRLFN_REGISTER, 0);

                        /* disable and enable endpoint to reset DATA PID */
                        ep_type = pdc_read32(EP_TYP_REGISTER);
                        ep_type &= (~EP_TYP_ENABLE);
                        pdc_write32(EP_TYP_REGISTER, ep_type);
                        pdc_write32(EP_TYP_REGISTER, (ep_type|EP_TYP_ENABLE) );

                        ep_type = pdc_read32(EP_TYP_REGISTER);

                        tx_data(pdc, handle, 2);
                    }
                    else 
                    {

                        pdc_write32(ENDPT_INDEX_REGISTER, handle);  
                        pdc_write32(CTRLFN_REGISTER, 0); 
                        /* disable and enable endpoint to reset DATA PID */
                        ep_type = pdc_read32(EP_TYP_REGISTER);
                        ep_type &= (~EP_TYP_ENABLE);
                        pdc_write32(EP_TYP_REGISTER, ep_type);
                        pdc_write32(EP_TYP_REGISTER, (ep_type|EP_TYP_ENABLE) );
                        rx_data(pdc, handle);
                    }
                }
                break;

            case PDC_GET_PIPE_STATUS:
                dev_check(" stall pipe status \n");
                pdc_write32(ENDPT_INDEX_REGISTER, handle);
                status = pdc_read32(CTRLFN_REGISTER);
                pipe_opr->pipe_status = (status&CNTRLFN_STALL) ? PDC_PIPE_STALL : PDC_PIPE_UNSTALL;

                pipe_opr->pipe_status = status; 
                break;
        }

        return 0;
    }

    return -1;
} /* End of pdc_pipe_operation() */

EXPORT_SYMBOL(pdc_pipe_operation);

/*--------------------------------------------------------------*
 *          Device Controller intialization functions
 *--------------------------------------------------------------*/
/* Initialisation of the Peripheral Controller */
void pdc_init(struct isp1761_dev        *dev)
{

    __u32 temp1;
    dev_print("++ pdc_init(dev=%p)\n",dev);

    /* set the ISP1761 Device controller Hardware Configuration */

     /*Disable all interrupts */
    pdc_write32(INT_ENABLE_REGISTER,0);

    /* Set default address */
    pdc_write32(ADDRESS_REGISTER,0);

    dev_info("pdc_init, chip id = 0x%08X \n", (pdc_read32(CHIPID_REGISTER)));
    pdc_write16(SCRATCH_REGISTER, 0xa5a5);
    temp1 = pdc_read16(SCRATCH_REGISTER);
/*    dev_info("pdc_init, scratch 0xa5a5 = 0x%08X \n", temp1);
    isp1761_reg_write16(isp1761_dc_dev,SCRATCH_REGISTER,0xc7c7);
    temp1 = pdc_read16(SCRATCH_REGISTER);    */
    dev_info("pdc_init, scratch 0xa5a5 = 0x%08X \n", temp1);
    return;
} /* End of pdc_init */



/* Actually connect to the bus */
void pdc_connect(void)
{
    /* Set device address & enable */
    pdc_write32(ADDRESS_REGISTER,ADDRESS_DEVEN|0);

    /*for device only  */
#ifndef CONFIG_USB_OTG
    pdc_write32(0x374,0x00000081);
    pdc_write32(0x374,0x00160000);
#endif
    /* Connect to the bus */
    pdc_write32(MODE_REGISTER, MODE_WKUPCS|MODE_CLKAON|MODE_GLINTENA);

    /*Default is set to 1us (0x1E). minimum pulse is 33.3ns */
    pdc_write32(INT_PULSE_SIZE_REG,DEFAULT_INT_PULSE_SIZE);


    /* set int config reg */
    pdc_write32(INT_CONFIG_REGISTER,DEFAULT_INT_CONFIG);

    /* Enable interrupts */
    pdc_write32(INT_ENABLE_REGISTER, IE_BASIC_CTRL_EP_MASK); 

#if 0
    /*To force the device to Full Speed mode */
    temp=pdc_read32(TESTMODE_REGISTER); 
    pdc_write32(TESTMODE_REGISTER,0x00000010); 
    temp=pdc_read32(TESTMODE_REGISTER); 
#endif

/*    temp=pdc_read32(ADDRESS_REGISTER);
    dev_info("pdc_connect: ADDRESS_REGISTER : 0x%x\n",temp);
    temp=pdc_read32(MODE_REGISTER);
    dev_info("pdc_connect: MODE_REGISTER : 0x%x\n",temp);

    temp=pdc_read32(INT_CONFIG_REGISTER);
    dev_info("pdc_connect: INT_CONFIG_REGISTER : 0x%x\n",temp);
    temp=pdc_read32(INT_ENABLE_REGISTER);
    dev_info("pdc_connect: INT_ENABLE_REGISTER : 0x%x\n",temp);
    temp=pdc_read32(HC_HW_MODE_REG);
    dev_info("pdc_connect: HC_MODECONTROL : 0x%x\n",temp);  */
    return;

} /* End of pdc_connect */

/* Disconnect bus */
void pdc_disconnect(void)
{
    int reg,temp;
    /* Reset device address & enable */
    pdc_write32(ADDRESS_REGISTER,0);

    /* Disable interrupts */
    pdc_write32(INT_ENABLE_REGISTER, 0); 

    /* Disconnect bus */
    reg=pdc_read32(MODE_REGISTER);
    reg|=0x10;
    pdc_write32(MODE_REGISTER, reg);
    reg&=~0x18;
    pdc_write32(MODE_REGISTER, reg);
    
    /*Turn off soft connect */
    pdc_write32(0x374,0x00010000);

/*    temp=pdc_read32(ADDRESS_REGISTER);
    printk("pdc_disconnect: ADDRESS_REGISTER : 0x%x\n",temp);
    temp=pdc_read32(MODE_REGISTER);
    printk("pdc_disconnect: MODE_REGISTER : 0x%x\n",temp);

    temp=pdc_read32(INT_CONFIG_REGISTER);
    printk("pdc_disconnect: INT_CONFIG_REGISTER : 0x%x\n",temp);
    temp=pdc_read32(INT_ENABLE_REGISTER);
    printk("pdc_disconnect: INT_ENABLE_REGISTER : 0x%x\n",temp);
    temp=pdc_read32(HC_HW_MODE_REG);
    printk("pdc_disconnect: HC_MODECONTROL : 0x%x\n",temp);  */
    return;

} /* End of pdc_disconnect */

EXPORT_SYMBOL(pdc_disconnect);


/*---------------------------------------------------------------------*
 *                   ISP1761 HOSAL interface functions                 *
 *---------------------------------------------------------------------*/

isp1761_id id = {
    .idVendor = 1234,
    .idProduct = 1234,
    .driver_info = 1234
};


struct isp1761_driver   pdc_driver = {
name:           "usb-pdc",
index:          ISP1761_DC,
id:          	&id,
probe:          pdc_probe,
remove:         pdc_remove,
};


/*Device initialisation */
int pdc_probe (struct isp1761_dev       *dev, isp1761_id *id) 
{
    struct pdc_dev *pdc=usb_devices;
    int result, i;
    struct pdc_pipe     *pipe;

    dev_print("++ pdc_probe(dev=%p)\n",dev);
    isp1761_dc_dev = dev;   
    /*for device only  */
#ifndef CONFIG_USB_OTG
//    pdc_write32(0x374,0x00060400);
    pdc_write32(0x374,0x00000081);
    pdc_write32(0x374,0x00160000);

    pdc_write32(0x374,0xFFFF0000); /* clear all otg interrupt */
    pdc_write32(0x374,0x00000481);
#endif
    pdc_write32(UNLOCK_DEVICE_REG, 0xaa37);

    pdc_write32(MODE_REGISTER, MODE_SFRESET);
    mdelay(1);
    pdc_write32(MODE_REGISTER, 0);
    mdelay(5);

    pdc->dev = dev;
    isp1761_dc_dev = dev;


    for(i=0; i< PDC_MAX_PIPES;i++) {
        pipe = pdc->ep_pipes + i;

        pipe->urb = NULL;
        pipe->txrx_idle = 1;
        pipe->ep_state = PDC_PIPE_UNCONFIG;
        pipe->ep_desc = NULL;
    }


    /* Configure the default end points 
       Configure the control OUT pipe */
    pipe = pdc->ep_pipes;
    pipe->ep_desc = &pdc_ctrl_ep_desc[0];

    pipe->ep_desc->ep_num = 0;
    pipe->ep_desc->ep_dir = PDC_EP_DIR_OUT;
    pipe->ep_desc->attributes = PDC_EP_CONTROL;
    pipe->ep_desc->max_pkt_size = 64;
    pipe->ep_state = PDC_PIPE_CONFIG;

    /* configure the control IN pipe */
    pipe = pdc->ep_pipes + 1;
    pipe->ep_desc = &pdc_ctrl_ep_desc[1];

    pipe->ep_desc->ep_num = 0;
    pipe->ep_desc->ep_dir = PDC_EP_DIR_IN;
    pipe->ep_desc->attributes = PDC_EP_CONTROL;
    pipe->ep_desc->max_pkt_size = 64;
    pipe->ep_state = PDC_PIPE_CONFIG;



    pdc_bus_init(pdc);

    /* Do chip setup */
    pdc_init(dev);

    /* Claim USB IRQ */
//    printk("MODE1:0x%08x\n",pdc_read32(MODE_REGISTER));
    result=isp1761_request_irq(pdc_isr,dev,pdc);
//    printk("MODE2:0x%08x\n",pdc_read32(MODE_REGISTER));

    /* Got it ok? */
    if (result < 0) {
        dev_check(KERN_ERR "usb-pdc:Can't get USB device IRQ %d.\n", dev->irq);
        return result;
    }
    dev_check("usb-pdc:device IRQ is initialised, Irq is  %04x\n",dev->irq);

    /* If there are pending IRQs, process them as we can only  detect edges */

#ifndef CONFIG_USB_OTG
    pdc_connect();
#endif 
    dev->driver_data = pdc;
    dev_check("pdc:Philips USB device driver in INIT.\n");
    return 0;

} /* End of pdc_probe */


void pdc_remove (struct isp1761_dev     *dev)  
{
    struct pdc_dev *pdc=(struct pdc_dev*)dev->driver_data;

    dev_print("++ pdc_remove(dev=%p)\n",dev);

    pdc_bus_deinit();

    /* Go off bus */
    pdc_write32(ADDRESS_REGISTER,0);

    /* Turn off IRQs */
    pdc_write32(INT_ENABLE_REGISTER,0);

    /* Global IRQ disable */
    pdc_write32(MODE_REGISTER,0);

    /*Turn off soft connect */
    pdc_write32(0x374,0x00010000);

    /* Free IRQ */
    isp1761_free_irq(dev,pdc);

    pdc->dev = NULL;
    isp1761_dc_dev = NULL;
    pdc->ep_pipes = pdc_eps;

} /* End of pdc_remove */

/*
 * This is module init function
 * This function registers the PDC driver to ISP1761 HOSAL driver
 * which in turn calls the probe function when the dev is found
 */
static int __init pdc_module_init (void) 
{
    struct pdc_dev  *pdc = usb_devices;

    int result;

    dev_print("++ pdc_module_init(void)\n");

    pdc->dev = NULL;
    pdc->ep_pipes = pdc_eps;
    isp1761_dc_dev = NULL;

    result = isp1761_register_driver(&pdc_driver);

    if(result == 0) {
        printk("ISP1761 PDC Device Initialization Success\n");
    } else {
        printk(KERN_INFO "Device Iinitialization Failed (error = %d)\n",result);
    }

    return  result;
} /* End of pdc_module_init */

/*
 * This is module cleanup function
 * Unregister from isp1761 HOSAL layer which in turn calls 
 * Close function
 */
static void __exit pdc_module_cleanup (void) 
{

    dev_print("++ pdc_module_cleanup (void)\n");

    return isp1761_unregister_driver(&pdc_driver);

} /* End of pdc_module_cleanup */


MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_LICENSE ("GPL");

module_init (pdc_module_init);
module_exit (pdc_module_cleanup);

